@@ -1,32 +1,34 @@
 import requests
 import sys
 from datetime import datetime
-from enum import StrEnum, auto
+from enum import StrEnum
 from logging import Logger
-from pypomes_core import TZ_LOCAL, exc_format
+from pypomes_core import (
+    APP_PREFIX, TZ_LOCAL, exc_format,
+    env_get_int, env_get_str, env_get_strs
+)
 from pypomes_crypto import crypto_jwk_convert
 from threading import RLock
 from typing import Any, Final
 
-
-class IamServer(StrEnum):
-    """
-    Supported IAM servers.
-    """
-    JUSBR = auto()
-    KEYCLOAK = auto()
+_members: dict[str, str] = {key.upper(): key.lower() for key in
+                            env_get_strs(key=f"{APP_PREFIX}_AUTH_SERVERS")}
+IamServer: type[StrEnum] = StrEnum("IamServer", _members)
+del _members
 
 
-class IamParam(StrEnum):
+class ServerParam(StrEnum):
     """
     Parameters for configuring *IAM* servers.
     """
+
     ADMIN_ID = "admin-id"
     ADMIN_SECRET = "admin-secret"
     CLIENT_ID = "client-id"
     CLIENT_REALM = "client-realm"
     CLIENT_SECRET = "client-secret"
     ENDPOINT_CALLBACK = "endpoint-callback"
+    ENDPOINT_CALLBACK_EXCHANGE = "endpoint-callback-exchange"
     ENDPOINT_LOGIN = "endpoint-login"
     ENDPOINT_LOGOUT = "endpoint_logout"
     ENDPOINT_TOKEN = "endpoint-token"
@@ -34,8 +36,9 @@ class IamParam(StrEnum):
     LOGIN_TIMEOUT = "login-timeout"
     PK_EXPIRATION = "pk-expiration"
     PK_LIFETIME = "pk-lifetime"
-    PUBLIC_KEY = "public-key"
     RECIPIENT_ATTR = "recipient-attr"
+    # dynamic attributes
+    PUBLIC_KEY = "public-key"
     URL_BASE = "url-base"
     USERS = "users"
 
@@ -54,31 +57,64 @@ class UserParam(StrEnum):
     REDIRECT_URI = "redirect-uri"
 
 
-# The configuration parameters for the IAM servers are specified dynamically dynamically with *iam_setup()*
-# Specifying configuration parameters with environment variables can be done in two ways:
-#
-# 1. for a single *IAM* server, specify the data set
-#       - *<APP_PREFIX>_IAM_ADMIN_ID*             (optional, needed if administrative duties are performed)
-#       - *<APP_PREFIX>_IAM_ADMIN_PWD*            (optional, needed if administrative duties are performed)
-#       - *<APP_PREFIX>_IAM_CLIENT_ID*            (required)
-#       - *<APP_PREFIX>_IAM_CLIENT_REALM*         (required)
-#       - *<APP_PREFIX>_IAM_CLIENT_SECRET*        (required)
-#       - *<APP_PREFIX>_IAM_ENDPOINT_CALLBACK*    (required)
-#       - *<APP_PREFIX>_IAM_ENDPOINT_EXCHANGE*    (required)
-#       - *<APP_PREFIX>_IAM_ENDPOINT_LOGIN*       (required)
-#       - *<APP_PREFIX>_IAM_ENDPOINT_LOGOUT*      (required)
-#       - *<APP_PREFIX>_IAM_ENDPOINT_PROVIDER*    (optional, needed if requesting tokens to providers)
-#       - *<APP_PREFIX>_IAM_ENDPOINT_TOKEN*       (required)
-#       - *<APP_PREFIX>_IAM_LOGIN_TIMEOUT*        (optional, defaults to no timeout)
-#       - *<APP_PREFIX>_IAM_PK_LIFETIME*          (optional, defaults to non-terminating lifetime)
-#       - *<APP_PREFIX>_IAM_RECIPIENT_ATTR*       (required)
-#       - *<APP_PREFIX>_IAM_URL_BASE*             (required)
-#
-# 2. for multiple *IAM* servers, specify the data set above for each server,
-#    respectively replacing *IAM* with a name in *IamServer* (currently, *JUSBR* and *KEYCLOAK* are supported).
-#
-# 3. the parameters *PUBLIC_KEY*, *PK_EXPIRATION*, and *USERS* cannot be assigned values,
-#    as they are reserved for internal use
+def __get_iam_data() -> dict[IamServer, dict[ServerParam, Any]]:
+    """
+    Obtain the configuration data for select *IAM* servers.
+
+    The configuration parameters for the IAM servers are specified dynamically with environment variables,
+    or dynamically with calls to *iam_setup_server()*. Specifying configuration parameters with environment
+    variables can be done by following these steps:
+
+    1. Specify *<APP_PREFIX>_AUTH_SERVERS* with a list of names among the values found in *IamServer* class
+       and the data set below for each server, where *<IAM>* stands for the server's name as presented in
+       *IamServer* class:
+          - *<APP_PREFIX>_<IAM>_ADMIN_ID*           (optional, required if administrative duties are performed)
+          - *<APP_PREFIX>_<IAM>_ADMIN_PWD*          (optional, required if administrative duties are performed)
+          - *<APP_PREFIX>_<IAM>_CLIENT_ID*          (required)
+          - *<APP_PREFIX>_<IAM>_CLIENT_REALM*       (required)
+          - *<APP_PREFIX>_<IAM>_CLIENT_SECRET*      (required)
+          - *<APP_PREFIX>_<IAM>_LOGIN_TIMEOUT*      (optional, defaults to no timeout)
+          - *<APP_PREFIX>_<IAM>_PK_LIFETIME*        (optional, defaults to non-terminating lifetime)
+          - *<APP_PREFIX>_<IAM>_RECIPIENT_ATTR*     (required)
+          - *<APP_PREFIX>_<IAM>_URL_BASE*           (required)
+
+    2. A group of special environment variables identifying endpoints for authentication services may be specified,
+       following the same scheme as presented in item *1* above. These are not part of the *IAM* server's setup,
+       but are meant to be used by function *iam_setup_endpoints()*, wherein the values in those variables
+       would represent default values for its parameters, respectively:
+          - *<APP_PREFIX>_<IAM>_ENDPOINT_CALLBACK*
+          - *<APP_PREFIX>_<IAM>_ENDPOINT_CALLBACK_EXCHANGE*
+          - *<APP_PREFIX>_<IAM>_ENDPOINT_EXCHANGE*
+          - *<APP_PREFIX>_<IAM>_ENDPOINT_LOGIN*
+          - *<APP_PREFIX>_<IAM>_ENDPOINT_LOGOUT*
+          - *<APP_PREFIX>_<IAM>_ENDPOINT_TOKEN*
+          - *<APP_PREFIX>_<IAM>_ENDPOINT_USERINFO*
+
+    :return: the configuration data for the select *IAM* servers.
+    """
+    # initialize the return variable
+    result: dict[IamServer, dict[ServerParam, Any]] = {}
+
+    for server in IamServer:
+        prefix = server.name
+        result[server] = {
+            ServerParam.ADMIN_ID: env_get_str(key=f"{APP_PREFIX}_{prefix}_ADMIN_ID"),
+            ServerParam.ADMIN_SECRET: env_get_str(key=f"{APP_PREFIX}_{prefix}_ADMIN_SECRET"),
+            ServerParam.CLIENT_ID: env_get_str(key=f"{APP_PREFIX}_{prefix}_CLIENT_ID"),
+            ServerParam.CLIENT_REALM: env_get_str(key=f"{APP_PREFIX}_{prefix}_CLIENT_REALM"),
+            ServerParam.CLIENT_SECRET: env_get_str(key=f"{APP_PREFIX}_{prefix}_CLIENT_SECRET"),
+            ServerParam.LOGIN_TIMEOUT: env_get_str(key=f"{APP_PREFIX}_{prefix}_LOGIN_TIMEOUT"),
+            ServerParam.PK_LIFETIME: env_get_int(key=f"{APP_PREFIX}_{prefix}_PK_LIFETIME"),
+            ServerParam.RECIPIENT_ATTR: env_get_str(key=f"{APP_PREFIX}_{prefix}_RECIPIENT_ATTR"),
+            ServerParam.URL_BASE: env_get_str(key=f"{APP_PREFIX}_{prefix}_URL_BASE"),
+            # dynamically set
+            ServerParam.PK_EXPIRATION: 0,
+            ServerParam.PUBLIC_KEY: None,
+            ServerParam.USERS: {}
+        }
+
+    return result
+
 
 # registry structure:
 # { <IamServer>:
@@ -91,6 +127,7 @@ class UserParam(StrEnum):
 #       "client-realm": <str,
 #       "client-timeout": <int>,
 #       "recipient-attr": <str>,
+#       # dynamic attributes
 #       "public-key": <str>,
 #       "pk-lifetime": <int>,
 #       "pk-expiration": <int>,
@@ -112,10 +149,10 @@ class UserParam(StrEnum):
 #   },
 #   ...
 # }
-_IAM_SERVERS: Final[dict[IamServer, dict[IamParam, Any]]] = {}
+_IAM_SERVERS: Final[dict[IamServer, dict[ServerParam, Any]]] = __get_iam_data()
 
 
-# the lock protecting the data in '_IAM_SERVERS'
+# the lock protecting the data in '_<IAM>_SERVERS'
 # (because it is 'Final' and set at declaration time, it can be accessed through simple imports)
 _iam_lock: Final[RLock] = RLock()
 
@@ -132,7 +169,7 @@ def _iam_server_from_endpoint(endpoint: str,
     :return: the corresponding *IAM* server, or *None* if one could not be obtained
     """
     # initialize the return variable
-    result: IamServer | None = None
+    result: type(IamServer) | None = None
 
     for iam_server in _IAM_SERVERS:
         if endpoint.startswith(iam_server):
@@ -161,12 +198,12 @@ def _iam_server_from_issuer(issuer: str,
     :return: the corresponding *IAM* server, or *None* if one could not be obtained
     """
     # initialize the return variable
-    result: IamServer | None = None
+    result: type(IamServer) | None = None
 
     for iam_server, registry in _IAM_SERVERS.items():
-        base_url: str = f"{registry[IamParam.URL_BASE]}/realms/{registry[IamParam.CLIENT_REALM]}"
+        base_url: str = f"{registry[ServerParam.URL_BASE]}/realms/{registry[ServerParam.CLIENT_REALM]}"
         if base_url == issuer:
-            result = IamServer(iam_server)
+            result = type(IamServer)(iam_server)
             break
 
     if not result:
@@ -227,9 +264,9 @@ def _get_public_key(iam_server: IamServer,
                                                  logger=logger)
     if registry:
         now: int = int(datetime.now(tz=TZ_LOCAL).timestamp())
-        if now > registry[IamParam.PK_EXPIRATION]:
+        if now > registry[ServerParam.PK_EXPIRATION]:
             # obtain the JWKS (JSON Web Key Set) from the token issuer
-            base_url: str = f"{registry[IamParam.URL_BASE]}/realms/{registry[IamParam.CLIENT_REALM]}"
+            base_url: str = f"{registry[ServerParam.URL_BASE]}/realms/{registry[ServerParam.CLIENT_REALM]}"
             url: str = f"{base_url}/protocol/openid-connect/certs"
             if logger:
                 logger.debug(msg=f"Obtaining signature public key used by IAM server '{iam_server}'")
@@ -251,9 +288,9 @@ def _get_public_key(iam_server: IamServer,
                         # convert from 'JWK' to 'PEM' and save it for further use
                         result = crypto_jwk_convert(jwk=jwk,
                                                     fmt="PEM")
-                        registry[IamParam.PUBLIC_KEY] = result
-                        lifetime: int = registry[IamParam.PK_LIFETIME] or 0
-                        registry[IamParam.PK_EXPIRATION] = now + lifetime if lifetime else sys.maxsize
+                        registry[ServerParam.PUBLIC_KEY] = result
+                        lifetime: int = registry[ServerParam.PK_LIFETIME] or 0
+                        registry[ServerParam.PK_EXPIRATION] = now + lifetime if lifetime else sys.maxsize
                         if logger:
                             logger.debug("Public key obtained and saved")
                     else:
@@ -278,7 +315,7 @@ def _get_public_key(iam_server: IamServer,
                 if isinstance(errors, list):
                     errors.append(msg)
         else:
-            result = registry[IamParam.PUBLIC_KEY]
+            result = registry[ServerParam.PUBLIC_KEY]
 
     return result
 
@@ -385,4 +422,4 @@ def _get_iam_users(iam_server: IamServer,
     registry: dict[str, Any] = _get_iam_registry(iam_server=iam_server,
                                                  errors=errors,
                                                  logger=logger)
-    return registry[IamParam.USERS] if registry else None
+    return registry[ServerParam.USERS] if registry else None
